#include <vector>
#include <string>
#include <fstream>
#include <sstream>

#include <boost/spirit/core.hpp>
#include <boost/spirit/attribute.hpp>
#include <boost/spirit/symbols/symbols.hpp>
#include <boost/spirit/utility/chset.hpp>
#include <boost/spirit/utility/escape_char.hpp>
#include <boost/spirit/utility/confix.hpp>
#include <boost/spirit/iterator.hpp>
#include <boost/spirit/error_handling/exceptions.hpp>

#include <boost/shared_ptr.hpp>
#include <boost/scoped_array.hpp>

#include <typeinfo>

#include "objfile.h"

// OBJ reader implementation code.
// Written by Christopher Twigg <cdtwigg@cs.cmu.edu>
// for 15-864 spring 2004.


using std::ifstream;
using std::string;
using namespace vl;
using namespace boost::spirit;

template <typename T>
class SetBool
{
public:
	SetBool( bool& valid, T& value )
		: value_(value), valid_(valid) {}

	void operator()( T value ) const
	{
		value_ = value;
		valid_ = true;
	}

private:
	T& value_;
	bool& valid_;
};

ObjFile::ObjFile( const std::string& filename )
	: filename_(filename)
{
	unsigned int numFaces = 0;

	const int maxline = 1000;
	std::ifstream ifs(filename.c_str());
	char line[maxline];

	std::cout << "Parsing .obj file '" << filename << "'." << std::endl;
	if( !ifs )
	{
		std::string message = "Couldn't open .obj file '";
		message.append( filename );
		message.append( "'" );
		throw ObjFileException( message );
	}

	int lineNum = 0;

	while( ifs )
	{
		++lineNum;

		ifs.getline( line, maxline );

		std::string command;
		parse_info<char const*> info = parse( line,
			lexeme_d[ +alpha_p[ append(command) ] ] );
		if( !info.hit )
			continue;

		const char* curPos = info.stop;

		if( command == "v" ) // vertex
		{
			vl::Vec3d pos;
			parse_info<char const*> info = parse( curPos, 
				real_p[assign(pos[0])] >> real_p[assign(pos[1])] >> real_p[assign(pos[2])],
				space_p );
			if( !info.full )
				throw ObjFileException("Invalid vertex", filename, lineNum, info.stop - line);

			vertexPositions_.push_back( pos );
		}
		else if( command == "vn" )
		{
			vl::Vec3d normal;
			parse_info<char const*> info = parse( curPos, 
				real_p[assign(normal[0])] >> real_p[assign(normal[1])] >> real_p[assign(normal[2])],
				space_p );
			if( !info.full )
				throw ObjFileException("Invalid normal", filename, lineNum, info.stop - line);

			normals_.push_back( normal );
		}
		else if( command == "vt" )
		{
			vl::Vec3d tex = vl_0;	// w defaults to 0
			parse_info<char const*> info = parse( curPos, 
				real_p[assign(tex[0])] >> real_p[assign(tex[1])] >> !(real_p[assign(tex[2])]),
				space_p );
			if( !info.full )
				throw ObjFileException("Invalid texture coordinate", filename, lineNum, info.stop - line);

			textureCoords_.push_back( tex );
		}
		else if( command == "g" )
		{
			std::string name;
			parse_info<char const*> info = parse( curPos, 
				*graph_p[ append(name) ], space_p );
			if( name.empty() )
				throw ObjFileException( "Empty group name", filename, lineNum, info.stop - line );				

			groups_.push_back( Group(name) );
		}
		else if( command == "f" )
		{
			if( groups_.empty() )
				groups_.push_back( Group("default") );

			Face face;

			const char* endLine = line + strlen(line);
			while( curPos < endLine )
			{
				unsigned int pos;
				std::pair< bool, unsigned int > texPos;
				std::pair< bool, unsigned int > normal;
				parse_info<char const*> info = parse( curPos,
					lexeme_d[ 
						int_p[ assign(pos) ] >> 
							!( ch_p('/') >> !int_p[SetBool<unsigned int>(texPos.first, texPos.second)]
							>> !( ch_p('/') >> !int_p[SetBool<unsigned int>(normal.first, normal.second)] ) ) ],
							space_p );

				if( !info.hit )
					throw ObjFileException( "Invalid vertex", filename, lineNum, info.stop - line );

				curPos = info.stop;
				face.addVertex( Vertex( pos, texPos, normal ) );
			}

			++numFaces;
			groups_.rbegin()->addFace( face );
		}
		else if( command == "mtllib" || command == "usemtl" || command == "s" )
		{
			// ignore for now
			std::cout << "Warning: ignoring '" << command << "' line" << std::endl;
		}
		else
		{
			std::ostringstream msg;
			msg << "Invalid line in .obj file '" << filename << "': " << line;
			throw ObjFileException(msg.str(), filename, lineNum, 1);
		}
	}

	// statistics
	std::cout << "Parsed obj file '" << filename << "'; statistics:" << std::endl;
	std::cout << "   " << groups_.size() << " groups," << std::endl;
	std::cout << "   " << numFaces << " faces," << std::endl;
	std::cout << "   " << vertexPositions_.size() << " vertices," << std::endl;
	std::cout << "   " << normals_.size() << " normals, " << std::endl;
	std::cout << "   " << textureCoords_.size() << " texture coordinates, " << std::endl;
}

std::vector<std::string> ObjFile::groupNames() const
{
	std::vector<std::string> result;
	result.reserve( groups_.size() );
	for( std::vector<Group>::const_iterator groupItr = groups_.begin();
		groupItr != groups_.end();
		++groupItr )
	{
		result.push_back( groupItr->name() );
	}

	return result;
}

ObjFile::Group ObjFile::group( std::string name ) const
{
	for( std::vector< Group >::const_iterator itr = groups_.begin();
		itr != groups_.end();
		++itr )
	{
		if( itr->name() == name )
			return *itr;
	}

	std::ostringstream oss;
	oss << "Invalid group name: '" << name << "'.";
	throw ObjFileException( oss.str() );
}

ObjFileException::ObjFileException( const std::string& reason, const std::string& filename,
		unsigned int line, unsigned int column )
{
	std::ostringstream oss;
	oss << "Error in file '" << filename 
		<< "', line " << line 
		<< ", column " << column << ": "
		<< reason;
	reason_ = oss.str();
}


