#ifndef __OBJFILE_H__

#define __OBJFILE_H__

#include <string>
#include <vector>
#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>
#include <vl/VLd.h>

// OBJ reader definitions.
// Written by Christopher Twigg <cdtwigg@cs.cmu.edu>
// for 15-864 spring 2004.


// thrown by the ObjFile constructor
class ObjFileException
{
public:
	ObjFileException( const std::string& reason )
		: reason_(reason) {}
	ObjFileException( const std::string& reason, const std::string& filename, 
		unsigned int line, unsigned int column );

	std::string reason() const { return reason_; }

private:
	std::string reason_;
};

// This is a wrapper for an .obj file.  We can construct one by simply
// passing in a filename; assuming the obj file does not have any errors
// we will then be able to access any of the vertices, etc. of the obj file.

// The .obj file format is documented fully here:
//    http://www.royriggs.com/obj.html
// A quick summary:
//   1.  vertices, normals, and texture coordinates are all specified in 
//       a global 1-based namespace.  
//   2.  faces are divided into groups
//   3.  each face consists of a listing of vertices, like this:
//          f 1/1/1 2/2/2 3/3/3
//       The numbers are references to the vertices, normals, and texture
//       coordinates, all of which were specified (as mentioned above) in
//       a global 1-based namespace.  
//
// So, with that said, to get an object out of the ObjFile object below
// once it has been constructed, do the following:
//   1.  Get the list of groups out of the .obj file using the "groupNames"
//       function.
//   2.  Select the group you want, and retrieve it using the "group" function.
//   3.  Iterate through the faces in the group using the "face" member function
//       of the Group class.
//   4.  Iterate through the vertices using the vertex member function of 
//       the Face class.
//   5.  Retrieve the various indices for the position(), textureCoordinate(), and 
//       normal() attributes of the vertex.
//   6.  Look these up in the .obj file global namspace using the vertexPosition,
//       textureCoordinate, and normal member functions of the ObjFile class.
class ObjFile
{
public:

	class Vertex
	{
	public:
		Vertex( const unsigned int& pos )
			:	position_(pos),
				texture_( std::make_pair( false, 0 ) ),
				normal_( std::make_pair( false, 0 ) ) {}

		Vertex( const unsigned int& pos, const unsigned int& texture )
			:	position_(pos),
				texture_( std::make_pair( true, texture ) ),
				normal_( std::make_pair( false, 0 ) ) {}
				
		Vertex( const unsigned int& pos, const unsigned int& texture, const unsigned int& normal )
			:	position_(pos),
				texture_( std::make_pair( true, texture ) ),
				normal_( std::make_pair( true, normal ) ) {}

		Vertex( const unsigned int pos, std::pair<bool, unsigned int> texture, std::pair<bool, unsigned int> normal )
			:	position_(pos),
				texture_(texture),
				normal_(normal) {}

		unsigned int position() const			{ return position_; }
		unsigned int normal() const				{ assert( normal_.first ); return normal_.second; }
		unsigned int textureCoordinate() const	{ assert( texture_.first ); return texture_.second; }

		// Normals and texture coordinates are not considered "required" in the
		// obj file format standard.  Check these before retrieving them.
		bool hasNormal() const	{ return normal_.first; }
		bool hasTextureCoordinate() const
								{ return texture_.first; }

	private:
		unsigned int position_; 
		std::pair< bool, unsigned int > texture_;
		std::pair< bool, unsigned int > normal_;	
	};

	class Face
	{
	public:
		Face() { vertices_.reserve( 3 ); }
		Face( const Vertex& v1, const Vertex& v2, const Vertex& v3 )
		{
			vertices_.reserve(3);
			vertices_.push_back(v1);
			vertices_.push_back(v2);
			vertices_.push_back(v3);
		}

	void addVertex( const Vertex& v )
		{
			vertices_.push_back( v );
		}

		unsigned int vertexCount() const { return vertices_.size(); }
		Vertex vertex(unsigned int vert) { return vertices_[vert]; }

	private:
		std::vector< Vertex > vertices_;
	};

	class Group
	{
	public:
		explicit Group( std::string name )
			: name_(name)
		{
		}

		void addFace( const Face& face )
		{
			faces_.push_back( face );
		}

		unsigned int faceCount() const { return faces_.size(); }
		Face face(unsigned int face) const { return faces_[face]; }
		std::string name() const { return name_; }

	private:
		std::string name_;
		std::vector< Face > faces_;
	};

	// Constructs the OBJ file and reads it in.  Throws an ObjFileException if
	// it fails for any reason (e.g., file not there, etc.).
	ObjFile( const std::string& filename );

	// Retrieve a group given a group name.
	Group group( std::string name ) const;

	// Retrieve a list of all the group names in the obj file.
	std::vector<std::string> groupNames() const;

	// obj files are 1-indexed
	vl::Vec3d vertexPosition( int iPos ) const { return vertexPositions_[iPos - 1]; }
	vl::Vec3d textureCoordinate( int iPos ) const { return textureCoords_[iPos - 1]; }
	vl::Vec3d normal( int iPos ) const { return normals_[iPos - 1]; }

private:
	std::vector< Group > groups_;
	std::vector< vl::Vec3d > vertexPositions_;
	std::vector< vl::Vec3d > textureCoords_;
	std::vector< vl::Vec3d > normals_;
	std::string filename_;
};

#endif
